@@ -8,31 +8,73 @@ use Exporter();
 *import = \&Exporter::import;
 our @EXPORT_OK = qw(
   urandom
+  urandom_ub
 );
+
 our %EXPORT_TAGS = ( 'all' => \@EXPORT_OK, );
 
-our $VERSION = '0.0.27';
+our @CARP_NOT = ('Crypt::URandom');
 
-my $CRYPT_SILENT      = hex '40';
-my $PROV_RSA_FULL     = 1;
-my $VERIFY_CONTEXT    = hex 'F0000000';
-my $W2K_MAJOR_VERSION = hex '5';
-my $W2K_MINOR_VERSION = 0;
+BEGIN {
+    our $VERSION = '0.41_01';
+    eval {
+        require XSLoader;
+
+        XSLoader::load( __PACKAGE__, $VERSION );
+    } or do {
+    };
+}
+
+## no critic (ProhibitConstantPragma)
+# using constant for the speed benefit of constant-folding of values
+
+use constant CRYPT_SILENT      => 64;               # hex 40
+use constant PROV_RSA_FULL     => 1;
+use constant VERIFY_CONTEXT    => 4_026_531_840;    # hex 'F0000000'
+use constant W2K_MAJOR_VERSION => 5;
+use constant W2K_MINOR_VERSION => 0;
+
+use constant OS_WIN32 => $OSNAME eq 'MSWin32';
+use constant PATH     => do {
+    my $path = '/dev/urandom';
+    if ( $OSNAME eq 'freebsd' ) {
+        $path = '/dev/random';    # FreeBSD's /dev/random is non-blocking
+    }
+    $path;
+};
+use constant GETRANDOM_AVAILABLE => do {
+    my $result = 0;
+    eval {
+        my $correct_length = 2;
+        my $actual_length =
+          crypt_urandom_getrandom( my $buffer = q[], $correct_length );
+        if ( $correct_length == $actual_length ) {
+            $result = 1;
+        }
+        $result;
+    } or do {
+        $result = undef;
+    };
+    $result;
+};
+
+## use critic
 
 my $_initialised;
 my $_context;
 my $_cryptgenrandom;
 my $_rtlgenrand;
+my $_urandom_handle;
 
 sub _init {
-    if ( $OSNAME eq 'MSWin32' ) {
+    if ( OS_WIN32() ) {
         require Win32;
         require Win32::API;
         require Win32::API::Type;
         my ( $major, $minor ) = ( Win32::GetOSVersion() )[ 1, 2 ];
-        my $ntorlower = ( $major < $W2K_MAJOR_VERSION ) ? 1 : 0;
+        my $ntorlower = ( $major < W2K_MAJOR_VERSION() ) ? 1 : 0;
         my $w2k =
-          ( $major == $W2K_MAJOR_VERSION and $minor == $W2K_MINOR_VERSION )
+          ( $major == W2K_MAJOR_VERSION() and $minor == W2K_MINOR_VERSION() )
           ? 1
           : 0;
 
@@ -53,8 +95,8 @@ sub _init {
 
             my $context = chr(0) x Win32::API::Type->sizeof('PULONG');
             my $result =
-              $crypt_acquire_context_a->Call( $context, 0, 0, $PROV_RSA_FULL,
-                $CRYPT_SILENT | $VERIFY_CONTEXT );
+              $crypt_acquire_context_a->Call( $context, 0, 0, PROV_RSA_FULL(),
+                CRYPT_SILENT() | VERIFY_CONTEXT() );
             my $pack_type = Win32::API::Type::packing('PULONG');
             $context = unpack $pack_type, $context;
             if ( !$result ) {
@@ -87,12 +129,26 @@ _RTLGENRANDOM_PROTO_
     }
     else {
         require FileHandle;
+        $_urandom_handle = FileHandle->new( PATH(), Fcntl::O_RDONLY() )
+          or Carp::croak(
+            q[Failed to open '] . PATH() . qq['. " for reading:$OS_ERROR] );
+        binmode $_urandom_handle;
     }
     return;
 }
 
+sub urandom_ub {
+    my ($length) = @_;
+    return _urandom( 'sysread', $length );
+}
+
 sub urandom {
     my ($length) = @_;
+    return _urandom( 'read', $length );
+}
+
+sub _urandom {
+    my ( $type, $length ) = @_;
 
     my $length_ok;
     if ( defined $length ) {
@@ -104,11 +160,15 @@ sub urandom {
         Carp::croak(
             'The length argument must be supplied and must be an integer');
     }
-    if ( !$_initialised ) {
-        _init();
-        $_initialised = 1;
+    if ( !GETRANDOM_AVAILABLE() ) {
+        if (
+            !( ( defined $_initialised ) && ( $_initialised == $PROCESS_ID ) ) )
+        {
+            _init();
+            $_initialised = $PROCESS_ID;
+        }
     }
-    if ( $OSNAME eq 'MSWin32' ) {
+    if ( OS_WIN32() ) {
         my $buffer = chr(0) x $length;
         if ($_cryptgenrandom) {
 
@@ -126,28 +186,33 @@ sub urandom {
         }
         return $buffer;
     }
-    else {
-        my ($path) = '/dev/urandom';
-        if ( $OSNAME eq 'freebsd' ) {
-            $path = '/dev/random';    # FreeBSD's /dev/random is non-blocking
+    elsif ( GETRANDOM_AVAILABLE() ) {
+        my $result = crypt_urandom_getrandom( my $buffer = q[], $length );
+        if ( $result == $length ) {
+            return $buffer;
         }
-        my ($handle) = FileHandle->new( $path, Fcntl::O_RDONLY() );
-        if ($handle) {
-            my ($result) = $handle->read( my $buffer, $length );
-            if ( defined $result ) {
-                if ( $result == $length ) {
-                    return $buffer;
-                }
-                else {
-                    Carp::croak("Only read $result bytes from '$path'");
-                }
+        else {
+            Carp::croak(qq[Only read $result bytes from getrandom]);
+        }
+    }
+    else {
+        my $result = $_urandom_handle->$type( my $buffer, $length );
+        if ( defined $result ) {
+            if ( $result == $length ) {
+                return $buffer;
             }
             else {
-                Carp::croak("Failed to read from '$path':$OS_ERROR");
+                $_urandom_handle = undef;
+                $_initialised    = undef;
+                Carp::croak(
+                    qq[Only read $result bytes from '] . PATH() . q['] );
             }
         }
         else {
-            Carp::croak("Failed to open '$path' for reading:$OS_ERROR");
+            my $error = $OS_ERROR;
+            $_urandom_handle = undef;
+            $_initialised    = undef;
+            Carp::croak( q[Failed to read from '] . PATH() . qq[':$error] );
         }
     }
 }
@@ -162,7 +227,7 @@ Crypt::URandom - Provide non blocking randomness
 
 =head1 VERSION
 
-This document describes Crypt::URandom version 0.0.27
+This document describes Crypt::URandom version 0.41_01
 
 
 =head1 SYNOPSIS
@@ -182,9 +247,8 @@ OR
 This Module is intended to provide
 an interface to the strongest available source of non-blocking 
 randomness on the current platform.  Platforms currently supported are
-anything supporting /dev/urandom and versions of Windows greater than 
-or equal to Windows 2000.
-
+anything supporting getrandom(2), /dev/urandom and versions of Windows greater
+than or equal to Windows 2000.
 
 =head1 SUBROUTINES/METHODS
 
@@ -192,9 +256,23 @@ or equal to Windows 2000.
 
 =item C<urandom>
 
+=for stopwords cryptographic
+
 This function accepts an integer and returns a string of the same size
-filled with random data.  The first call will initialise the native 
-cryptographic libraries (if necessary) and load all the required Perl libraries
+filled with random data.  The first call will initialize the native
+cryptographic libraries (if necessary) and load all the required Perl libraries.
+This call is a buffered read on non Win32 platforms that do not support getrandom(2)
+or equivalent.
+
+=item C<urandom_ub>
+
+=for stopwords cryptographic
+
+This function accepts an integer and returns a string of the same size
+filled with random data.  The first call will initialize the native
+cryptographic libraries (if necessary) and load all the required Perl libraries.
+This call is a unbuffered sysread on non Win32 platforms that do not support
+getrandom(2) or equivalent.
 
 =back
 
@@ -209,21 +287,29 @@ cryptographic functions provided by the operating system.
 
 =item C<Could not import CryptAcquireContext>
 
+=for stopwords CryptAcquireContextA advapi32
+
 The module was unable to load the CryptAcquireContextA function from the 
-advapi32 dynamic library.  The advapi32library cannot probably be loaded.
+advapi32 dynamic library.  The advapi32 library cannot probably be loaded.
 
 =item C<CryptAcquireContext failed>
 
+=for stopwords advapi32
+
 The module was unable to call the CryptAcquireContextA function from the
 advapi32 dynamic library.
 
 =item C<Could not import CryptGenRandom>
 
+=for stopwords advapi32 CryptGenRandom
+
 The module was unable to load the CryptGenRandom function from the 
 advapi32 dynamic library.
 
 =item C<Could not import SystemFunction036>
 
+=for stopwords SystemFunction036
+
 The module was unable to load the SystemFunction036 function from the 
 advapi32 dynamic library.
 
@@ -239,6 +325,8 @@ amount of randomness
 
 =item C<RtlGenRand failed>
 
+=for stopwords RtlGenRand
+
 The post Windows 2000 RtlGenRand method call failed to generate the required
 amount of randomness
 
@@ -265,6 +353,8 @@ Crypt::URandom requires no configuration files or environment variables.
 
 =over
 
+=for stopwords perl
+
 If the platform is Win32, the Win32::API module will be required.  Otherwise
 no other modules other than those provided by perl will be required
 
@@ -277,33 +367,33 @@ None reported.
 
 =head1 BUGS AND LIMITATIONS
 
-No bugs have been reported.
-
-Please report any bugs or feature requests to
-C<bug-crypt-urandom@rt.cpan.org>, or through the web interface at
-L<http://rt.cpan.org>.
-
+To report a bug, or view the current list of bugs, please visit L<https://github.com/david-dick/crypt-urandom/issues>
 
 =head1 AUTHOR
 
 David Dick  C<< <ddick@cpan.org> >>
 
+=for stopwords ACKNOWLEDGEMENTS
 
 =head1 ACKNOWLEDGEMENTS
 
+=for stopwords CryptoAPI Kanat-Alexander
+
 The Win32::API code for interacting with Microsoft's CryptoAPI was stolen with extreme
 gratitude from Crypt::Random::Source::Strong::Win32 by Max Kanat-Alexander
 
 =head1 LICENSE AND COPYRIGHT
 
-Copyright (c) 2011, David Dick C<< <ddick@cpan.org> >>. All rights reserved.
+Copyright (c) 2023, David Dick C<< <ddick@cpan.org> >>. All rights reserved.
 
 This module is free software; you can redistribute it and/or
-modify it under the same terms as Perl itself. See L<perlartistic>.
+modify it under the same terms as Perl itself.
 
 
 =head1 DISCLAIMER OF WARRANTY
 
+=for stopwords MERCHANTABILITY LICENCE
+
 BECAUSE THIS SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
 FOR THE SOFTWARE, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT WHEN
 OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
